@@ -12,7 +12,7 @@
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
  *
- * $Id: pkcs11_lib.c 271 2007-05-21 08:19:49Z ludovic.rousseau $
+ * $Id: pkcs11_lib.c 379 2009-12-03 08:40:22Z ludovic.rousseau $
  */
 
 #define __PKCS11_LIB_C__
@@ -27,6 +27,7 @@
 #include <unistd.h>
 #include <errno.h>
 #include <string.h>
+#include <ctype.h>
 #include "debug.h"
 #include "error.h"
 #include "cert_info.h"
@@ -67,13 +68,76 @@ int pkcs11_pass_login(pkcs11_handle_t *h, int nullok)
   return 0;
 }
 
+/*
+ * memcmp_pad_max() is a specialized version of memcmp() which compares two
+ * pieces of data up to a maximum length.  If the two data match up the
+ * maximum length, they are considered matching.  Trailing blanks do not cause
+ * the match to fail if one of the data is shorted.
+ *
+ * Examples of matches:
+ *	"one"           |
+ *	"one      "     |
+ *	                ^maximum length
+ *
+ *	"Number One     |  X"	(X is beyond maximum length)
+ *	"Number One   " |
+ *	                ^maximum length
+ *
+ * Examples of mismatches:
+ *	" one"
+ *	"one"
+ *
+ *	"Number One    X|"
+ *	"Number One     |"
+ *	                ^maximum length
+ */
+static int
+memcmp_pad_max(void *d1, size_t d1_len, void *d2, size_t d2_len,
+    size_t max_sz)
+{
+	size_t		len, extra_len;
+	char		*marker;
+
+	/* No point in comparing anything beyond max_sz */
+	if (d1_len > max_sz)
+		d1_len = max_sz;
+	if (d2_len > max_sz)
+		d2_len = max_sz;
+
+	/* Find shorter of the two data. */
+	if (d1_len <= d2_len) {
+		len = d1_len;
+		extra_len = d2_len;
+		marker = d2;
+	} else {	/* d1_len > d2_len */
+		len = d2_len;
+		extra_len = d1_len;
+		marker = d1;
+	}
+
+	/* Have a match in the shortest length of data? */
+	if (memcmp(d1, d2, len) != 0)
+		/* CONSTCOND */
+		return (1);
+
+	/* If the rest of longer data is nulls or blanks, call it a match. */
+	while (len < extra_len && marker[len])
+		if (!isspace(marker[len++]))
+			/* CONSTCOND */
+			return (1);
+	return (0);
+}
+
+
 #ifdef HAVE_NSS
 /*
  * Using NSS to find the manage the PKCS #11 modules
  */
 #include "nss.h"
+#include "nspr.h"
 #include "cert.h"
 #include "secmod.h"
+#include "secutil.h"
 #include "pk11pub.h"
 #include "cert_st.h"
 #include "secutil.h"
@@ -100,7 +164,7 @@ struct pkcs11_handle_str {
 static int app_has_NSS = 0;
 
 
-char *
+static char *
 password_passthrough(PK11SlotInfo *slot, PRBool retry, void *arg)
 {
   /* give up if 1) no password was supplied, or 2) the password has already
@@ -148,7 +212,7 @@ int crypto_init(cert_policy *policy) {
 }
 
 
-static SECMODModule *find_module_by_library(char *pkcs11_module) 
+static SECMODModule *find_module_by_library(char *pkcs11_module)
 {
   SECMODModule *module = NULL;
   SECMODModuleList *modList = SECMOD_GetDefaultModuleList();
@@ -171,7 +235,7 @@ static SECMODModule *find_module_by_library(char *pkcs11_module)
 /*
  * NSS allows you to load a specific module. If the user specified a module
  * to load, load it, otherwize select on of the standard modules from the
- * secmod.db list. 
+ * secmod.db list.
  */
 int load_pkcs11_module(char *pkcs11_module, pkcs11_handle_t **hp)
 {
@@ -196,7 +260,7 @@ int load_pkcs11_module(char *pkcs11_module, pkcs11_handle_t **hp)
     return 0;
   }
 
-  /* specified module is not already loaded, load it now */	
+  /* specified module is not already loaded, load it now */
   moduleSpec = (char *)malloc(sizeof(SPEC_TEMPLATE) + strlen(pkcs11_module));
   if (!moduleSpec) {
     DBG1("Malloc failed when allocating module spec", strerror(errno));
@@ -228,12 +292,12 @@ int init_pkcs11_module(pkcs11_handle_t *h, int flag)
   return 0; /* NSS initialized the module on load */
 }
 
-int find_slot_by_number(pkcs11_handle_t *h, int slot_num, unsigned int *slotID)
+int find_slot_by_number(pkcs11_handle_t *h, unsigned int slot_num, unsigned int *slotID)
 {
   SECMODModule *module = h->module;
   int i;
 
-  /* if module is null, 
+  /* if module is null,
    * any of the PKCS #11 modules specified in the system config
    * is available, find one */
   if (module == NULL) {
@@ -286,9 +350,9 @@ int find_slot_by_number(pkcs11_handle_t *h, int slot_num, unsigned int *slotID)
   }
   /* we're configured for a specific module and token, see if it's present */
   slot_num--;
-  if (slot_num >= 0 && slot_num < module->slotCount && module->slots &&
-      module->slots[i] && PK11_IsPresent(module->slots[i])) {
-    h->slot = PK11_ReferenceSlot(module->slots[i]);
+  if (slot_num < module->slotCount && module->slots &&
+      module->slots[slot_num] && PK11_IsPresent(module->slots[slot_num])) {
+    h->slot = PK11_ReferenceSlot(module->slots[slot_num]);
     *slotID = PK11_GetSlotID(h->slot);
     return 0;
   }
@@ -299,36 +363,36 @@ int find_slot_by_number(pkcs11_handle_t *h, int slot_num, unsigned int *slotID)
  * find a slot by it's slot number or label. If slot number is '0' any
  * slot is ok.
  */
-int find_slot_by_number_and_label(pkcs11_handle_t *h, 
+int find_slot_by_number_and_label(pkcs11_handle_t *h,
 				  int wanted_slot_id,
-				  const char *wanted_slot_label,
+				  const char *wanted_token_label,
                                   unsigned int *slot_num)
 {
   int rv;
-  const char *slot_label = NULL;
+  const char *token_label = NULL;
   PK11SlotInfo *slot = NULL;
 
   /* we want a specific slot id, or we don't kare about the label */
-  if ((wanted_slot_label == NULL) || (wanted_slot_id != 0)) {
+  if ((wanted_token_label == NULL) || (wanted_slot_id != 0)) {
     rv = find_slot_by_number(h, wanted_slot_id, slot_num);
 
     /* if we don't care about the label, or we failed, we're done */
-    if ((wanted_slot_label == NULL) || (rv != 0)) {
+    if ((wanted_token_label == NULL) || (rv != 0)) {
       return rv;
     }
 
     /* verify it's the label we want */
-    slot_label = PK11_GetTokenName(h->slot);
+    token_label = PK11_GetTokenName(h->slot);
 
-    if ((slot_label != NULL) && 
-        (strcmp (wanted_slot_label, slot_label) == 0)) {
+    if ((token_label != NULL) &&
+        (strcmp (wanted_token_label, token_label) == 0)) {
       return 0;
     }
     return -1;
   }
 
   /* we want a specific slot by label only */
-  slot = PK11_FindSlotByName(wanted_slot_label);
+  slot = PK11_FindSlotByName(wanted_token_label);
   if (!slot) {
     return -1;
   }
@@ -348,9 +412,9 @@ int find_slot_by_number_and_label(pkcs11_handle_t *h,
   return 0;
 }
 
-int wait_for_token(pkcs11_handle_t *h, 
+int wait_for_token(pkcs11_handle_t *h,
                    int wanted_slot_id,
-                   const char *wanted_slot_label,
+                   const char *wanted_token_label,
                    unsigned int *slot_num)
 {
   int rv;
@@ -358,7 +422,7 @@ int wait_for_token(pkcs11_handle_t *h,
   rv = -1;
   do {
     /* see if the card we're looking for is inserted */
-    rv = find_slot_by_number_and_label (h, wanted_slot_id,  wanted_slot_label,
+    rv = find_slot_by_number_and_label (h, wanted_slot_id,  wanted_token_label,
                                         slot_num);
     if (rv !=  0) {
       PK11SlotInfo *slot;
@@ -385,7 +449,150 @@ int wait_for_token(pkcs11_handle_t *h,
   return rv;
 }
 
-void release_pkcs11_module(pkcs11_handle_t *h) 
+/*
+ * This function will search the slot list to find a slot based on the slot
+ * label.  If the wanted_slot_label is "none", then we will return the first
+ * slot with the token presented.
+ *
+ * This function return 0 if it found a matching slot; otherwise, it returns
+ * -1.
+ */
+int
+find_slot_by_slotlabel(pkcs11_handle_t *h, const char *wanted_slot_label,
+    unsigned int *slotID)
+{
+  SECMODModule *module = h->module;
+  PK11SlotInfo *slot;
+  int rv;
+  int i;
+
+  if (slotID == NULL || wanted_slot_label == NULL ||
+      strlen(wanted_slot_label) == 0 || module == NULL)
+    return (-1);
+
+  if (strcmp(wanted_slot_label, "none") == 0) {
+    rv = find_slot_by_number(h, 0, slotID);
+    return (rv);
+  } else {
+    /* wanted_slot_label is not "none"  */
+    for (i = 0; i < module->slotCount; i++) {
+      if (module->slots[i] && PK11_IsPresent(module->slots[i])) {
+        const char *slot_label;
+
+	slot = PK11_ReferenceSlot(module->slots[i]);
+	slot_label = PK11_GetSlotName(slot);
+	if (memcmp_pad_max((void *)slot_label, strlen(slot_label),
+	    (void *)wanted_slot_label, strlen(wanted_slot_label), 64) == 0) {
+	  h->slot = slot;
+	  *slotID = PK11_GetSlotID(slot);
+	  return 0;
+        }
+      }
+    }
+  }
+
+  return (-1);
+}
+
+
+int
+find_slot_by_slotlabel_and_tokenlabel(pkcs11_handle_t *h,
+    const char *wanted_slot_label, const char *wanted_token_label,
+    unsigned int *slot_num)
+{
+  SECMODModule *module = h->module;
+  PK11SlotInfo *slot;
+  unsigned long i;
+  int rv;
+
+  if (slot_num == NULL || module == NULL)
+    return (-1);
+
+  if (wanted_token_label == NULL){
+    rv = find_slot_by_slotlabel(h, wanted_slot_label, slot_num);
+    return (rv);
+  }
+
+  /* wanted_token_label != NULL */
+  if (strcmp(wanted_slot_label, "none") == 0) {
+    for (i = 0; i < module->slotCount; i++) {
+      if (module->slots[i] && PK11_IsPresent(module->slots[i])) {
+	const char *token_label;
+	slot = PK11_ReferenceSlot(module->slots[i]);
+	token_label = PK11_GetTokenName(slot);
+	if (memcmp_pad_max((void *) token_label, strlen(token_label),
+	    (void *)wanted_token_label, strlen(wanted_token_label), 33) == 0) {
+          h->slot =  slot;
+          *slot_num =  PK11_GetSlotID(slot);
+	  return (0);
+        }
+      }
+    }
+    return (-1);
+  } else {
+    for (i = 0; i < module->slotCount; i++) {
+      if (module->slots[i] && PK11_IsPresent(module->slots[i])) {
+	const char *slot_label;
+	const char *token_label;
+
+	slot = PK11_ReferenceSlot(module->slots[i]);
+	slot_label = PK11_GetSlotName(slot);
+	token_label = PK11_GetTokenName(slot);
+	if ((memcmp_pad_max((void *)slot_label, strlen(slot_label),
+	    (void *)wanted_slot_label, strlen(wanted_slot_label), 64) == 0) &&
+            (memcmp_pad_max((void *)token_label, strlen(token_label),
+	    (void *)wanted_token_label, strlen(wanted_token_label), 33) == 0))
+	{
+          h->slot =  slot;
+          *slot_num =  PK11_GetSlotID(slot);
+	  return (0);
+        }
+      }
+    }
+    return (-1);
+  }
+}
+
+int wait_for_token_by_slotlabel(pkcs11_handle_t *h,
+                   const char *wanted_slot_label,
+                   const char *wanted_token_label,
+                   unsigned int *slot_num)
+{
+  int rv;
+
+  rv = -1;
+  do {
+    /* see if the card we're looking for is inserted */
+    rv = find_slot_by_slotlabel_and_tokenlabel (h, wanted_slot_label,
+	wanted_token_label, slot_num);
+
+    if (rv !=  0) {
+      PK11SlotInfo *slot;
+      PRIntervalTime slot_poll_interval; /* only for legacy hardware */
+
+      /* if the card is not inserted, then block until something happens */
+      slot_poll_interval = PR_MillisecondsToInterval(PAM_PKCS11_POLL_TIME);
+      slot = SECMOD_WaitForAnyTokenEvent(h->module, 0 /* flags */,
+                                         slot_poll_interval);
+
+      /* unexpected error */
+      if (slot == NULL) {
+        break;
+      }
+
+      /* something happened, continue loop and check if the card
+       * we're looking for is inserted
+       */
+      PK11_FreeSlot(slot);
+      continue;
+    }
+  } while (rv != 0);
+
+  return rv;
+}
+
+
+void release_pkcs11_module(pkcs11_handle_t *h)
 {
   SECStatus rv;
   close_pkcs11_session(h);
@@ -470,7 +677,7 @@ int close_pkcs11_session(pkcs11_handle_t *h)
   return 0;
 }
 
-const char *get_slot_label(pkcs11_handle_t *h)
+const char *get_slot_tokenlabel(pkcs11_handle_t *h)
 {
   if (!h->slot) {
     return NULL;
@@ -507,7 +714,7 @@ cert_object_t **get_certificate_list(pkcs11_handle_t *h, int *count)
   rv = CERT_FilterCertListByUsage(certList,  certUsageSSLClient, PR_FALSE);
   if (rv != SECSuccess) {
       CERT_DestroyCertList(certList);
-      DBG1("Couldn't filter out email certs: %s", 
+      DBG1("Couldn't filter out email certs: %s",
 				SECU_Strerror(PR_GetError()));
       return NULL;
   }
@@ -521,7 +728,7 @@ cert_object_t **get_certificate_list(pkcs11_handle_t *h, int *count)
   }
 
   /* convert the link list from NSS to the array used by pam_pkcs11 */
-  for (node = CERT_LIST_HEAD(certList); !CERT_LIST_END(node,certList); 
+  for (node = CERT_LIST_HEAD(certList); !CERT_LIST_END(node,certList);
 						node = CERT_LIST_NEXT(node)) {
 	if (node->cert) {
 	    DBG3("cert %d: found (%s), \"%s\"", certCount,
@@ -541,7 +748,7 @@ cert_object_t **get_certificate_list(pkcs11_handle_t *h, int *count)
     return NULL;
   }
 
-  for (node = CERT_LIST_HEAD(certList); !CERT_LIST_END(node,certList); 
+  for (node = CERT_LIST_HEAD(certList); !CERT_LIST_END(node,certList);
                                          node = CERT_LIST_NEXT(node)) {
     if (node->cert) {
       certs[certIndex++] = (cert_object_t *)CERT_DupCertificate(node->cert);
@@ -604,7 +811,7 @@ int sign_value(pkcs11_handle_t *h, cert_object_t *cert, CK_BYTE *data,
   return 0;
 }
 
-int get_random_value(unsigned char *data, int length) 
+int get_random_value(unsigned char *data, int length)
 {
   SECStatus rv = PK11_GenerateRandom(data,length);
   if (rv != SECSuccess) {
@@ -613,7 +820,6 @@ int get_random_value(unsigned char *data, int length)
   return (rv == SECSuccess) ? 0 : -1;
 }
 
-#include "nspr.h"
 
 struct tuple_str {
     PRErrorCode	 errNum;
@@ -641,7 +847,7 @@ const PRInt32 numStrings = sizeof(errStrings) / sizeof(tuple_str);
  * Returns NULL of errNum is unknown.
  */
 const char *
-SECU_Strerror(PRErrorCode errNum) 
+SECU_Strerror(PRErrorCode errNum)
 {
   PRInt32 low  = 0;
   PRInt32 high = numStrings - 1;
@@ -657,12 +863,12 @@ SECU_Strerror(PRErrorCode errNum)
     for (i = low; i <= high; ++i) {
       num = errStrings[i].errNum;
       if (num <= lastNum) {
-        fprintf(stderr, 
+        fprintf(stderr,
                 "sequence error in error strings at item %d\n"
                 "error %d (%s)\n"
                 "should come after \n"
                 "error %d (%s)\n",
-                i, lastNum, errStrings[i-1].errString, 
+                i, lastNum, errStrings[i-1].errString,
                 num, errStrings[i].errString);
       }
       lastNum = num;
@@ -674,11 +880,11 @@ SECU_Strerror(PRErrorCode errNum)
   while (low + 1 < high) {
     i = (low + high) / 2;
     num = errStrings[i].errNum;
-    if (errNum == num) 
+    if (errNum == num)
       return errStrings[i].errString;
     if (errNum < num)
       high = i;
-    else 
+    else
       low = i;
   }
   if (errNum == errStrings[low].errNum)
@@ -708,7 +914,8 @@ struct cert_object_str {
 typedef struct {
   CK_SLOT_ID id;
   CK_BBOOL token_present;
-  CK_UTF8CHAR label[33];
+  CK_UTF8CHAR label[33]; /* token label */
+  CK_UTF8CHAR slotDescription[64];
 } slot_t;
 
 struct pkcs11_handle_str {
@@ -741,7 +948,7 @@ int load_pkcs11_module(char *module, pkcs11_handle_t **hp)
 
   DBG1("PKCS #11 module = [%s]", module);
   /* reset pkcs #11 handle */
-  
+
   h = (pkcs11_handle_t *)calloc(sizeof(pkcs11_handle_t), 1);
   if (h == NULL) {
     set_error("pkcs11_handle_t malloc failed: %s", strerror(errno));
@@ -758,7 +965,7 @@ int load_pkcs11_module(char *module, pkcs11_handle_t **hp)
   DBG3("module permissions: uid = %d, gid = %d, mode = %o",
       module_stat.st_uid, module_stat.st_gid, module_stat.st_mode & 0777);
   if (module_stat.st_mode & S_IWGRP || module_stat.st_mode & S_IWOTH
-      || module_stat.st_uid != 0 || module_stat.st_gid != 0) {
+      || module_stat.st_uid != 0) {
     set_error("the pkcs #11 module MUST be owned by root and MUST NOT "
               "be writeable by the group or others");
     free(h);
@@ -782,7 +989,7 @@ int load_pkcs11_module(char *module, pkcs11_handle_t **hp)
   }
   rv = C_GetFunctionList_ptr(&h->fl);
   if (rv != CKR_OK) {
-    set_error("C_GetFunctionList() failed: %x", rv);
+    set_error("C_GetFunctionList() failed: 0x%08lX", rv);
     free(h);
     return -1;
   }
@@ -804,9 +1011,13 @@ refresh_slots(pkcs11_handle_t *h)
     DBG1("slot %d:", i + 1);
     rv = h->fl->C_GetSlotInfo(h->slots[i].id, &sinfo);
     if (rv != CKR_OK) {
-      set_error("C_GetSlotInfo() failed: %x", rv);
+      set_error("C_GetSlotInfo() failed: 0x%08lX", rv);
       return -1;
     }
+
+    (void) memcpy(h->slots[i].slotDescription, sinfo.slotDescription,
+		sizeof(h->slots[i].slotDescription));
+
     DBG1("- description: %.64s", sinfo.slotDescription);
     DBG1("- manufacturer: %.32s", sinfo.manufacturerID);
     DBG1("- flags: %04lx", sinfo.flags);
@@ -814,7 +1025,7 @@ refresh_slots(pkcs11_handle_t *h)
       DBG("- token:");
       rv = h->fl->C_GetTokenInfo(h->slots[i].id, &tinfo);
       if (rv != CKR_OK) {
-        set_error("C_GetTokenInfo() failed: %x", rv);
+        set_error("C_GetTokenInfo() failed: 0x%08lX", rv);
         return -1;
       }
       DBG1("  - label: %.32s", tinfo.label);
@@ -837,8 +1048,8 @@ int init_pkcs11_module(pkcs11_handle_t *h,int flag)
   CK_SLOT_ID_PTR slots;
   CK_INFO info;
   CK_C_INITIALIZE_ARGS initArgs;
-  /* 
-   Set up arguments to allow native threads 
+  /*
+   Set up arguments to allow native threads
    According with pkcs#11v2.20, must set all pointers to null
    and flags CKF_OS_LOCKING_OK
   */
@@ -855,13 +1066,13 @@ int init_pkcs11_module(pkcs11_handle_t *h,int flag)
   if (rv == CKR_OK)
     h->should_finalize = 1;
   else if (rv != CKR_CRYPTOKI_ALREADY_INITIALIZED) {
-    set_error("C_Initialize() failed: %x", rv);
+    set_error("C_Initialize() failed: 0x%08lX", rv);
     return -1;
   }
 
   rv = h->fl->C_GetInfo(&info);
   if (rv != CKR_OK) {
-    set_error("C_GetInfo() failed: %x", rv);
+    set_error("C_GetInfo() failed: 0x%08lX", rv);
     return -1;
   }
   /* show some information about the module */
@@ -874,7 +1085,7 @@ int init_pkcs11_module(pkcs11_handle_t *h,int flag)
   /* get a list of all slots */
   rv = h->fl->C_GetSlotList(FALSE, NULL, &h->slot_count);
   if (rv != CKR_OK) {
-    set_error("C_GetSlotList() failed: %x", rv);
+    set_error("C_GetSlotList() failed: 0x%08lX", rv);
     return -1;
   }
   DBG1("number of slots (a): %d", h->slot_count);
@@ -897,7 +1108,7 @@ int init_pkcs11_module(pkcs11_handle_t *h,int flag)
   rv = h->fl->C_GetSlotList(FALSE, slots, &h->slot_count);
   if (rv != CKR_OK) {
     free(slots);
-    set_error("C_GetSlotList() failed: %x", rv);
+    set_error("C_GetSlotList() failed: 0x%08lX", rv);
     return -1;
   }
   DBG1("number of slots (b): %d", h->slot_count);
@@ -925,11 +1136,11 @@ void release_pkcs11_module(pkcs11_handle_t *h)
   free(h);
 }
 
-int find_slot_by_number(pkcs11_handle_t *h, int slot_num, unsigned int *slot)
+int find_slot_by_number(pkcs11_handle_t *h, unsigned int slot_num, unsigned int *slot)
 {
    /* zero means find the best slot */
    if (slot_num == 0) {
-	for (slot_num = 0; slot_num < h->slot_count && 
+	for (slot_num = 0; slot_num < h->slot_count &&
 				!h->slots[slot_num].token_present; slot_num++);
    } else {
 	/* otherwize it's an index into the slot table  (it is *NOT* the slot
@@ -942,30 +1153,30 @@ int find_slot_by_number(pkcs11_handle_t *h, int slot_num, unsigned int *slot)
    *slot = slot_num;
    return 0;
 }
-	
-int find_slot_by_number_and_label(pkcs11_handle_t *h, 
+
+int find_slot_by_number_and_label(pkcs11_handle_t *h,
 				  int wanted_slot_id,
-				  const char *wanted_slot_label,
+				  const char *wanted_token_label,
                                   unsigned int *slot_num)
 {
   unsigned int slot_index;
   int rv;
-  const char *slot_label = NULL;
+  const char *token_label = NULL;
 
   /* we want a specific slot id, or we don't care about the label */
-  if ((wanted_slot_label == NULL) || (wanted_slot_id != 0)) {
+  if ((wanted_token_label == NULL) || (wanted_slot_id != 0)) {
     rv = find_slot_by_number(h, wanted_slot_id, slot_num);
 
     /* if we don't care about the label, or we failed, we're done */
-    if ((wanted_slot_label == NULL) || (rv != 0)) {
+    if ((wanted_token_label == NULL) || (rv != 0)) {
       return rv;
     }
 
     /* verify it's the label we want */
-    slot_label = h->slots[*slot_num].label;
+    token_label = h->slots[*slot_num].label;
 
-    if ((slot_label != NULL) && 
-        (strcmp (wanted_slot_label, slot_label) == 0)) {
+    if ((token_label != NULL) &&
+        (strcmp (wanted_token_label, token_label) == 0)) {
       return 0;
     }
     return -1;
@@ -974,9 +1185,9 @@ int find_slot_by_number_and_label(pkcs11_handle_t *h,
   /* look up the slot by it's label from the list */
   for (slot_index = 0; slot_index < h->slot_count; slot_index++) {
     if (h->slots[slot_index].token_present) {
-      slot_label = h->slots[slot_index].label;
-      if ((slot_label != NULL) && 
-          (strcmp (wanted_slot_label, slot_label) == 0)) {
+      token_label = h->slots[slot_index].label;
+      if ((token_label != NULL) &&
+          (strcmp (wanted_token_label, token_label) == 0)) {
         *slot_num = slot_index;
         return 0;
       }
@@ -985,9 +1196,99 @@ int find_slot_by_number_and_label(pkcs11_handle_t *h,
   return -1;
 }
 
-int wait_for_token(pkcs11_handle_t *h, 
-                   int wanted_slot_id,
+
+/*
+ * This function will search the slot list to find a slot based on the slot
+ * label.  If the wanted_slot_label is "none", then we will return the first
+ * slot with the token presented.
+ *
+ * This function return 0 if it found a matching slot; otherwise, it returns
+ * -1.
+ */
+int
+find_slot_by_slotlabel(pkcs11_handle_t *h, const char *wanted_slot_label,
+    unsigned int *slot_num)
+{
+  unsigned long idx;
+  size_t len;
+
+  if (slot_num == NULL || wanted_slot_label == NULL ||
+      strlen(wanted_slot_label) == 0)
+    return (-1);
+
+  if (strcmp(wanted_slot_label, "none") == 0) {
+    for (idx = 0; idx < h->slot_count; idx++) {
+      if (h->slots[idx].token_present) {
+	*slot_num = idx;
+	return (0);
+      }
+    }
+  } else {
+    /* Look up the slot by it's slotDescription */
+    len = strlen(wanted_slot_label);
+    for (idx = 0; idx < h->slot_count; idx++) {
+      if (h->slots[idx].token_present
+	  && memcmp_pad_max(h->slots[idx].slotDescription, 64,
+	  (void *)wanted_slot_label, len, 64) == 0) {
+	*slot_num = idx;
+	return (0);
+      }
+    }
+  }
+
+  return (-1);
+}
+
+
+int
+find_slot_by_slotlabel_and_tokenlabel(pkcs11_handle_t *h,
+    const char *wanted_slot_label, const char *wanted_token_label,
+    unsigned int *slot_num)
+{
+  unsigned long i;
+  int rv;
+
+  if (slot_num == NULL)
+    return (-1);
+
+  if (wanted_token_label == NULL) {
+    rv = find_slot_by_slotlabel(h, wanted_slot_label, slot_num);
+    return (rv);
+  }
+
+  /* wanted_token_label != NULL */
+  if (strcmp(wanted_slot_label, "none") == 0) {
+    for (i= 0; i < h->slot_count; i++) {
+      if (h->slots[i].token_present &&
+	  strcmp(wanted_token_label, h->slots[i].label) == 0) {
+	*slot_num = i;
+	return (0);
+      }
+    }
+    return (-1);
+  } else {
+    for (i = 0; i < h->slot_count; i++) {
+      if (h->slots[i].token_present) {
+        const char *slot_label = h->slots[i].slotDescription;
+        const char *token_label = h->slots[i].label;
+
+	if ((memcmp_pad_max((void *)slot_label, strlen(slot_label),
+	    (void *)wanted_slot_label, strlen(wanted_slot_label), 64) == 0) &&
+            (memcmp_pad_max((void *)token_label, strlen(token_label),
+            (void *)wanted_token_label, strlen(wanted_token_label), 33) == 0))
+        {
+	  *slot_num = i;
+	  return (0);
+        }
+      }
+    }
+    return (-1);
+  }
+}
+
+int wait_for_token_by_slotlabel(pkcs11_handle_t *h,
                    const char *wanted_slot_label,
+                   const char *wanted_token_label,
                    unsigned int *slot_num)
 {
   int rv;
@@ -995,7 +1296,30 @@ int wait_for_token(pkcs11_handle_t *h,
   rv = -1;
   do {
     /* see if the card we're looking for is inserted */
-    rv = find_slot_by_number_and_label (h, wanted_slot_id,  wanted_slot_label,
+    rv = find_slot_by_slotlabel_and_tokenlabel (h, wanted_slot_label,
+	wanted_token_label, slot_num);
+    if (rv !=  0) {
+      /* could call C_WaitForSlotEvent, for now just poll */
+      sleep(10);
+      refresh_slots(h);
+      continue;
+    }
+  } while (rv != 0);
+
+  return rv;
+}
+
+int wait_for_token(pkcs11_handle_t *h,
+                   int wanted_slot_id,
+                   const char *wanted_token_label,
+                   unsigned int *slot_num)
+{
+  int rv;
+
+  rv = -1;
+  do {
+    /* see if the card we're looking for is inserted */
+    rv = find_slot_by_number_and_label (h, wanted_slot_id,  wanted_token_label,
                                         slot_num);
     if (rv !=  0) {
       /* could call C_WaitForSlotEvent, for now just poll */
@@ -1016,11 +1340,11 @@ int open_pkcs11_session(pkcs11_handle_t *h, unsigned int slot)
   if (slot >= h->slot_count) {
     set_error("invalid slot number %d", slot);
     return -1;
-  } 
+  }
   /* open a readonly user-session */
   rv = h->fl->C_OpenSession(h->slots[slot].id, CKF_SERIAL_SESSION, NULL, NULL, &h->session);
   if (rv != CKR_OK) {
-    set_error("C_OpenSession() failed: %x", rv);
+    set_error("C_OpenSession() failed: 0x%08lX", rv);
     return -1;
   }
   h->current_slot = slot;
@@ -1034,12 +1358,25 @@ int pkcs11_login(pkcs11_handle_t *h, char *password)
   DBG("login as user CKU_USER");
   rv = h->fl->C_Login(h->session, CKU_USER, (unsigned char*)password, strlen(password));
   if (rv != CKR_OK) {
-    set_error("C_Login() failed: %x", rv);
+    set_error("C_Login() failed: 0x%08lX", rv);
     return -1;
   }
   return 0;
 }
 
+int get_slot_login_required(pkcs11_handle_t *h)
+{
+  int rv;
+  CK_TOKEN_INFO tinfo;
+
+  rv = h->fl->C_GetTokenInfo(h->slots[h->current_slot].id, &tinfo);
+  if (rv != CKR_OK) {
+    set_error("C_GetTokenInfo() failed: 0x%08lX", rv);
+    return -1;
+  }
+  return tinfo.flags & CKF_LOGIN_REQUIRED;
+}
+
 static void free_certs(cert_object_t **certs, int cert_count)
 {
   int i;
@@ -1065,13 +1402,13 @@ int close_pkcs11_session(pkcs11_handle_t *h)
   DBG("logout user");
   rv = h->fl->C_Logout(h->session);
   if (rv != CKR_OK && rv != CKR_USER_NOT_LOGGED_IN) {
-    set_error("C_Logout() failed: %x", rv);
+    set_error("C_Logout() failed: 0x%08lX", rv);
     return -1;
   }
   DBG("closing the PKCS #11 session");
   rv = h->fl->C_CloseSession(h->session);
   if (rv != CKR_OK) {
-    set_error("C_CloseSession() failed: %x", rv);
+    set_error("C_CloseSession() failed: 0x%08lX", rv);
     return -1;
   }
   DBG("releasing keys and certificates");
@@ -1084,7 +1421,7 @@ int close_pkcs11_session(pkcs11_handle_t *h)
 }
 
 /* get a list of certificates */
-cert_object_t **get_certificate_list(pkcs11_handle_t *h, int *ncerts) 
+cert_object_t **get_certificate_list(pkcs11_handle_t *h, int *ncerts)
 {
   CK_BYTE *id_value;
   CK_BYTE *cert_value;
@@ -1093,7 +1430,7 @@ cert_object_t **get_certificate_list(pkcs11_handle_t *h, int *ncerts)
   X509 *x509;
   cert_object_t **certs = NULL;
   int rv;
-  
+
   CK_OBJECT_CLASS cert_class = CKO_CERTIFICATE;
   CK_CERTIFICATE_TYPE cert_type = CKC_X_509;
   CK_ATTRIBUTE cert_template[] = {
@@ -1113,14 +1450,14 @@ cert_object_t **get_certificate_list(pkcs11_handle_t *h, int *ncerts)
 
   rv = h->fl->C_FindObjectsInit(h->session, cert_template, 2);
   if (rv != CKR_OK) {
-    set_error("C_FindObjectsInit() failed: %x", rv);
+    set_error("C_FindObjectsInit() failed: 0x%08lX", rv);
     return NULL;
   }
   while(1) {
     /* look for certificates */
     rv = h->fl->C_FindObjects(h->session, &object, 1, &object_count);
     if (rv != CKR_OK) {
-      set_error("C_FindObjects() failed: %x", rv);
+      set_error("C_FindObjects() failed: 0x%08lX", rv);
       goto getlist_error;
     }
     if (object_count == 0) break; /* no more certs */
@@ -1134,7 +1471,7 @@ cert_object_t **get_certificate_list(pkcs11_handle_t *h, int *ncerts)
     cert_template[2].ulValueLen = 0;
     rv = h->fl->C_GetAttributeValue(h->session, object, cert_template, 3);
     if (rv != CKR_OK) {
-        set_error("CertID length: C_GetAttributeValue() failed: %x", rv);
+        set_error("CertID length: C_GetAttributeValue() failed: 0x%08lX", rv);
         goto getlist_error;
     }
     /* allocate enought space */
@@ -1148,7 +1485,7 @@ cert_object_t **get_certificate_list(pkcs11_handle_t *h, int *ncerts)
     rv = h->fl->C_GetAttributeValue(h->session, object, cert_template, 3);
     if (rv != CKR_OK) {
         free(id_value);
-        set_error("CertID value: C_GetAttributeValue() failed: %x", rv);
+        set_error("CertID value: C_GetAttributeValue() failed: 0x%08lX", rv);
         goto getlist_error;
     }
 
@@ -1158,7 +1495,7 @@ cert_object_t **get_certificate_list(pkcs11_handle_t *h, int *ncerts)
       cert_template[3].pValue = NULL;
       rv = h->fl->C_GetAttributeValue(h->session, object, cert_template, 4);
       if (rv != CKR_OK) {
-        set_error("Cert Length: C_GetAttributeValue() failed: %x", rv);
+        set_error("Cert Length: C_GetAttributeValue() failed: 0x%08lX", rv);
         goto getlist_error;
       }
     /* allocate enought space */
@@ -1172,7 +1509,7 @@ cert_object_t **get_certificate_list(pkcs11_handle_t *h, int *ncerts)
       rv = h->fl->C_GetAttributeValue(h->session, object, cert_template, 4);
       if (rv != CKR_OK) {
         free(cert_value);
-        set_error("Cert Value: C_GetAttributeValue() failed: %x", rv);
+        set_error("Cert Value: C_GetAttributeValue() failed: 0x%08lX", rv);
         goto getlist_error;
       }
 
@@ -1219,7 +1556,7 @@ cert_object_t **get_certificate_list(pkcs11_handle_t *h, int *ncerts)
   /* release FindObject Sesion */
   rv = h->fl->C_FindObjectsFinal(h->session);
   if (rv != CKR_OK) {
-    set_error("C_FindObjectsFinal() failed: %x", rv);
+    set_error("C_FindObjectsFinal() failed: 0x%08lX", rv);
     free_certs(certs, h->cert_count);
     certs = NULL;
     h->certs = NULL;
@@ -1237,7 +1574,7 @@ cert_object_t **get_certificate_list(pkcs11_handle_t *h, int *ncerts)
 getlist_error:
   rv = h->fl->C_FindObjectsFinal(h->session);
   if (rv != CKR_OK) {
-    set_error("C_FindObjectsFinal() failed: %x", rv);
+    set_error("C_FindObjectsFinal() failed: 0x%08lX", rv);
   }
   free_certs(h->certs, h->cert_count);
   h->certs = NULL;
@@ -1272,24 +1609,24 @@ int get_private_key(pkcs11_handle_t *h, cert_object_t *cert) {
   key_template[3].ulValueLen = cert->id_length;
   rv = h->fl->C_FindObjectsInit(h->session, key_template, 2);
   if (rv != CKR_OK) {
-    set_error("C_FindObjectsInit() failed: %x", rv);
+    set_error("C_FindObjectsInit() failed: 0x%08lX", rv);
     return -1;
   }
   rv = h->fl->C_FindObjects(h->session, &object, 1, &object_count);
   if (rv != CKR_OK) {
-    set_error("C_FindObjects() failed: %x", rv);
+    set_error("C_FindObjects() failed: 0x%08lX", rv);
     goto get_privkey_failed;
   }
   if (object_count <= 0) {
       /* cert without prk: perhaps CA or CA-chain cert */
-      set_error("No private key found for cert: %x", rv);
+      set_error("No private key found for cert: 0x%08lX", rv);
       goto get_privkey_failed;
   }
 
   /* and finally release Find session */
   rv = h->fl->C_FindObjectsFinal(h->session);
   if (rv != CKR_OK) {
-    set_error("C_FindObjectsFinal() failed: %x", rv);
+    set_error("C_FindObjectsFinal() failed: 0x%08lX", rv);
     return -1;
   }
 
@@ -1301,12 +1638,12 @@ int get_private_key(pkcs11_handle_t *h, cert_object_t *cert) {
 get_privkey_failed:
   rv = h->fl->C_FindObjectsFinal(h->session);
   if (rv != CKR_OK) {
-    set_error("C_FindObjectsFinal() failed: %x", rv);
+    set_error("C_FindObjectsFinal() failed: 0x%08lX", rv);
   }
   return -1;
 }
 
-const char *get_slot_label(pkcs11_handle_t *h)
+const char *get_slot_tokenlabel(pkcs11_handle_t *h)
 {
   return h->slots[h->current_slot].label;
 }
@@ -1316,7 +1653,7 @@ const X509 *get_X509_certificate(cert_object_t *cert)
   return cert->x509;
 }
 
-int sign_value(pkcs11_handle_t *h, cert_object_t *cert, CK_BYTE *data, 
+int sign_value(pkcs11_handle_t *h, cert_object_t *cert, CK_BYTE *data,
 	CK_ULONG length, CK_BYTE **signature, CK_ULONG *signature_length)
 {
   int rv;
@@ -1328,7 +1665,7 @@ int sign_value(pkcs11_handle_t *h, cert_object_t *cert, CK_BYTE *data,
   if (get_private_key(h, cert) == -1) {
     set_error("Couldn't find private key for certificate");
     return -1;
-  } 
+  }
 
   /* set mechanism */
   switch (cert->key_type) {
@@ -1346,7 +1683,7 @@ int sign_value(pkcs11_handle_t *h, cert_object_t *cert, CK_BYTE *data,
   /* sign the token */
   rv = h->fl->C_SignInit(h->session, &mechanism, cert->private_key);
   if (rv != CKR_OK) {
-    set_error("C_SignInit() failed: %x", rv);
+    set_error("C_SignInit() failed: 0x%08lX", rv);
     return -1;
   }
   *signature = NULL;
@@ -1367,7 +1704,7 @@ int sign_value(pkcs11_handle_t *h, cert_object_t *cert, CK_BYTE *data,
     } else if (rv != CKR_OK) {
       free(*signature);
       *signature = NULL;
-      set_error("C_Sign() failed: %x", rv);
+      set_error("C_Sign() failed: 0x%08lX", rv);
       return -1;
     }
   }
